///////////////////Software Control Header///////////////////////////////////
// JPGLib.h - Library to support JPG encoding/decoding
//
// $LogThisFile$ \Projects\UtilCode\Include\JPGLib.h
// 
// V01.01 04/11/2002 2223
//  Minor updates.
// 
// V01.00 06/10/2002 1049
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#ifndef __JPG_LIB__
#define __JPG_LIB__

// Define _TS_DEBUG_ in your project if you want to dump out (cout) messages
// during decoding of JPG bitstream.

#ifdef _TS_DEBUG_
#include <iostream.h>
#endif

#include "BitStrm.h"

// Constants
// Start of frame markers, non-differential, Huffman coding
#define JPG_SOF0	0xFFC0	// Baseline DCT
#define JPG_SOF1	0xFFC1	// Extended sequential DCT
#define JPG_SOF2	0xFFC2	// Progressive DCT
#define JPG_SOF3	0xFFC3	// Lossless (sequential)

// Start of frame markers, differential, Huffman coding
#define JPG_SOF5	0xFFC5	// Differential sequential DCT
#define JPG_SOF6	0xFFC6	// Differential progressive DCT
#define JPG_SOF7	0xFFC7	// Differential lossless (sequential)

// Start of frame markers, non-differential, arithmetic coding
#define JPG_JPG	0xFFC8	// Reserved for JPEG extensions
#define JPG_SOF9	0xFFC9	// Extended sequential DCT
#define JPG_SOF10	0xFFCA	// Progressive DCT
#define JPG_SOF11	0xFFCB	// Lossless (sequential)

// Start of frame markers, differential, arithmetic coding
#define JPG_SOF13	0xFFCD	// Differential sequential DCT
#define JPG_SOF14 0xFFCE	// Differential progressive DCT
#define JPG_SOF15	0xFFCF	// Differential lossless (sequential)

// Huffman table specification
#define JPG_DHT	0xFFC4	// Define Huffman table(s)

// Arithmetic coding conditioning specification
#define JPG_DAC	0xFFCC	// Define arithmetic coding conditioning(s)

// Restart interval termination
#define JPG_RST_BASE	0xFFD0	// Restart with modula 8 count "m"

// Other markers
#define JPG_SOI	0xFFD8	// Start of image
#define JPG_EOI	0xFFD9	// End of image
#define JPG_SOS	0xFFDA	// Start of scan
#define JPG_DQT	0xFFDB	// Define quantization table(s)
#define JPG_DNL	0xFFDC	// Define number of lines
#define JPG_DRI	0xFFDD	// Define restart interval
#define JPG_DHP	0xFFDE	// Define hierarchical progression
#define JPG_EXP	0xFFDF	// Expand reference component(s)
#define JPG_APP_BASE	0xFFE0	// Reserved for application segments
#define JPG_JPG_BASE	0xFFF0	// Reserved for JPEG extensions
#define JPG_COM	0xFFFE	// Comment

// Reserved markers
#define JPG_TEM	0xFF01	// For temporary private use in arithmetic coding
#define JPG_RES_BASE	0xFF02	// Reserved

#define PI 3.141592654

/////////////////////////////////////////////////////////////////////////////
// CJPG_FrameHeader
//
// Desctiption: Object to hold the frame header.
//
// TSTan   06 Oct 2002
/////////////////////////////////////////////////////////////////////////////
class CJPG_FrameHeader
{
public:
	int m_type, m_precision, m_y, m_x, m_nf;
	char *m_imgcomp;

	CJPG_FrameHeader();
	~CJPG_FrameHeader();

///////////////////
// Data retrieval
///////////////////
	int GetC(int index);
	int GetH(int index);
	int GetTq(int index);
	int GetV(int index);

///////////////////
// Modifiers
///////////////////
	void Reset();
};

/////////////////////////////////////////////////////////////////////////////
// CJPG_HuffTable
//
// Desctiption: Object holding Huffman table.
//
// TSTan   06 Oct 2002
/////////////////////////////////////////////////////////////////////////////
class CJPG_HuffTable
{
public:
	int m_tableclass;
	int m_length[16];
	char *m_codes;

	// Generated tables
	int m_huffcode[256], m_huffsize[256];
	int m_mincode[16], m_maxcode[16], m_valptr[16];

	CJPG_HuffTable();
	~CJPG_HuffTable();

///////////////////
// Modifiers
///////////////////
	void Reset();
};

/////////////////////////////////////////////////////////////////////////////
// CJPGEngine
//
// Desctiption: Engine to decode/encode JPG bitstream.
//
// TSTan   06 Oct 2002
/////////////////////////////////////////////////////////////////////////////
class CJPGEngine
{
	CBitStream *m_iobits;
	CJPG_FrameHeader m_frmheader;
	CJPG_HuffTable m_huff[4];

	int m_quantize[4][64];
	char *m_scanentropy;
	char m_scan_ss, m_scan_se, m_scan_ah, m_scan_al;

	// Decoder use
	int m_dec_qe, m_dec_sense;

	// Decoder/Encoder use
	int m_a, m_b, m_b2, m_c, m_ct, m_cnt, m_lastk;

	// DCT
	int m_zz[64], m_pred[3];
	int *m_data, *m_pixels;

	// Image info
	char *m_img_type, *m_img_thumb, *m_img;
	int m_img_ver_major, m_img_ver_minor, m_img_units;
	int m_img_x, m_img_y, m_img_xthumb, m_img_ythumb;

	// Restart interval
	int m_restart, m_status;

	////////////////////////////////////////////////////////
	//                Decoder                             //
	////////////////////////////////////////////////////////
	int Decode(int hufftable);
	int DecodeDU(int *dataptr, int component);
	int DecodeScan();
	void Decode_ZZ(int k, int ssss);
	int GetUpSample(int largeX, int largeY, int thisX, int thixY,
									 int lookX, int lookY, int *dataptr);
	void LosslessDecode();	// Unimplemented
	void ProgressiveDecode();	// Unimplemented
	int NextBit();
	void SequentialDecode(int *dataptr, int component);

	// Marker processing
	int InterpreteMarker(int SinglePass = 0);
	void ProcessAPP(int type);
	void ProcessDHT(int skiplength = 0);
	void ProcessDQT(int skiplength = 0);
	int ProcessDRI();
	void ProcessSOF(int token);
	int ProcessSOS();

	// Huffman tables
	void GenCodeTable(int *phuffsize, int *phuffcode);
	void GenSizeTable(int *pbits, int *phuffsize);

	////////////////////////////////////////////////////////
	//                Encoder                             //
	////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////
	//                Shared                              //
	////////////////////////////////////////////////////////
	void InitDec();
	void OrderCodes(int *phuffval, int *phuffcode, int *phuffsize,
						 int *pehufco, int *pehufsi);

	void ByteIn();
	void Unstuff_0();

public:
	CJPGEngine();
	~CJPGEngine();

///////////////////
// Data retrieval
///////////////////
	// Return pointer to RGB buffer storing image
	char *GetImage() {return m_img;}
	// Return X pixel count
	int GetX() {return m_frmheader.m_x;}
	// Return Y pixel count
	int GetY() {return m_frmheader.m_y;}

///////////////////
// Modifiers
///////////////////
	void ProcessInFile(char *filename);
	void Reset();
};

#endif
