///////////////////Software Control Header///////////////////////////////////
// BitStrm.cpp - Library to bit IO operation.
//
// $LogThisFile$ \Projects\UtilCode\Lib\BitStrm.cpp
// 
// V01.01 02/11/2002 2326
//  File seeking functions added.
// 
// V01.00 06/03/2001 2254
//  Initial revision.
// 
// This code is freely distributed to help the development of complex
// software in any field using C++. The author is not held responsible
// for any damage caused by the usage of the code.
// The author should be acknowledged for the use of this piece of code in
// commercial software.
//
// Bug reports and suggestions are welcomed to report to
//
//          tzzstan@hotmail.com
/////////////////////////////////////////////////////////////////////////////
#include "BitStrm.h"

/////////////////////////////////////////////////////////////////////////////
// CBufferBlock()
//
// Desctiption: Allocate empty memory block.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
CBufferBlock::CBufferBlock()
{
   m_Data = new char[BUFFER_LENGTH];
   m_ByteIndex = 0;
   m_BitIndex = 7;
}

/////////////////////////////////////////////////////////////////////////////
// CBufferBlock(CBufferBlock &source)
//
// Desctiption: Make a copy of source to new memory area.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
CBufferBlock::CBufferBlock(CBufferBlock &source)
{
   m_Data = new char[BUFFER_LENGTH];
   m_ByteIndex = 0;
   m_BitIndex = 7;
   memcpy(m_Data, source.m_Data, source.SizeInBytes());
}

/////////////////////////////////////////////////////////////////////////////
// ~CBufferBlock()
//
// Desctiption: Cleanup.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
CBufferBlock::~CBufferBlock()
{
   delete []m_Data;
}

/////////////////////////////////////////////////////////////////////////////
// AddBits(int data, int bits)
//
// Desctiption: Add bits to output stream.
//
// data - Data bits right aligned with last bit at bit 0.
// bits - Bits to be added.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
int CBufferBlock::AddBits(int data, int bits)
{
   int AllowedBits;

   AllowedBits = (SpaceInBits() < bits)? SpaceInBits(): bits;
   int Remains = bits - AllowedBits;
   int AllowedData = data >> Remains;

	// Loop until less than 8 bit
   while(AllowedBits >= 8)
   {
      if(m_BitIndex == 7)	// If byte aligned
      {
         int theData = AllowedData >> (AllowedBits - 8);
         m_Data[m_ByteIndex] = theData & 0x00ff;
         m_ByteIndex++;
			AllowedBits -= 8;
      }
      else
      {
         // simplified from (AllowedBits - 8 + (7 - m_BitIndex))
         // Upper portion
         int theData = AllowedData >> (AllowedBits - 1 - m_BitIndex);
         int Mask = 0x00fe << m_BitIndex;
         m_Data[m_ByteIndex] &= (Mask & 0x00ff);
         Mask = ~Mask & 0x00ff;
         m_Data[m_ByteIndex] |= (theData & Mask);
         m_ByteIndex++;
			AllowedBits -= m_BitIndex + 1;
			m_BitIndex = 7;
      }
   }

   if(AllowedBits >= (m_BitIndex + 1))
   {
      int theData = AllowedData >> (AllowedBits - m_BitIndex - 1);
      int Mask = 0x00fe << m_BitIndex;
      m_Data[m_ByteIndex] &= (Mask & 0x00ff);
      Mask = ~Mask & 0x00ff;
      m_Data[m_ByteIndex] |= (theData & Mask);
      AllowedBits -= m_BitIndex + 1;
      m_ByteIndex++;
      m_BitIndex = 7;
   }

   if(AllowedBits)
   {
      // Simplified from (8 - AllowedBits) - (7 - m_BitIndex)
      int theData = AllowedData << (m_BitIndex + 1 - AllowedBits);
      int Mask = 0x00fe << m_BitIndex;
      m_Data[m_ByteIndex] &= Mask;
      m_Data[m_ByteIndex] |= (theData & ~Mask);
      m_BitIndex -= AllowedBits;
      if(m_BitIndex < 0)
         m_BitIndex = 7;
   }

   return Remains;
}

/////////////////////////////////////////////////////////////////////////////
// SizeInBytes()
//
// Desctiption: Bytes of memory used.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
int CBufferBlock::SizeInBytes()
{
   int Length = m_ByteIndex;
   if(m_BitIndex != 7)
      Length++;

   return Length;
}

/////////////////////////////////////////////////////////////////////////////
// SpaceInBits()
//
// Desctiption: Remaining free bits.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
int CBufferBlock::SpaceInBits()
{
   int Length;

   Length = (BUFFER_LENGTH - m_ByteIndex) * 8;
   Length -= (7 - m_BitIndex);

   return Length;
}

/////////////////////////////////////////////////////////////////////////////
// CBitStream()
//
// Desctiption: Initiate new IO bit stream.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
CBitStream::CBitStream()
{
   Reset();
}

/////////////////////////////////////////////////////////////////////////////
// ~CBitStream()
//
// Desctiption: Cleanup.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
CBitStream::~CBitStream()
{
}

/////////////////////////////////////////////////////////////////////////////
// AddBits(CTString &BitString)
//
// Desctiption: Add bits to output stream.
//
// BitString - String containing bits to output. First bit at rightmost
//             location.
//                       "01000100010"
//                                  ^- 1st bit
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::AddBits(CTString &BitString)
{
   int iBit, StrLength, BitStr, BitIdx, BitLength;

   BitLength = sizeof(int) << 3;
   StrLength = BitString.Size();
   BitStr = 0;
   BitIdx = 0;
   for(iBit = StrLength - 1; iBit >= 0; iBit--)
   {
      BitStr <<= 1;
      BitStr |= BitString[iBit] == '1'? 1: 0;
      BitIdx++;

      if(BitIdx == BitLength)
      {
         AddBits(BitStr, BitIdx);
         BitStr = 0;
         BitIdx = 0;
      }
   }

   if(BitIdx)
      AddBits(BitStr, BitIdx);
}

/////////////////////////////////////////////////////////////////////////////
// AddBits(int data, int bits)
//
// Desctiption: Add data bits to output stream.
//              Initiate new block if overflow.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::AddBits(int data, int bits)
{
   if(m_ActiveOBuffer == NULL)
   {
      m_ActiveOBuffer = new CBufferBlock;
      m_OutputBuffers.Insert(m_ActiveOBuffer, CSinglyList<CBufferBlock>::TAIL);
   }

   int RemainBits = m_ActiveOBuffer->AddBits(data, bits);
   if(RemainBits)
   {
      m_ActiveOBuffer = new CBufferBlock;
      m_OutputBuffers.Insert(m_ActiveOBuffer, CSinglyList<CBufferBlock>::TAIL);
      m_ActiveOBuffer->AddBits(data, RemainBits);
   }
}

/////////////////////////////////////////////////////////////////////////////
// FillArray(char *array, int size)
//
// Desctiption: Get bytes from the bitstream and fill into array.
//
// array - pointer to array to fill up.
// size - size of array.
//
// TSTan   20 Oct 2002
/////////////////////////////////////////////////////////////////////////////
void CBitStream::FillArray(char *array, int size)
{
	while(size--)
	{
		*array = GetBits(8);
		array++;
	}
}

/////////////////////////////////////////////////////////////////////////////
// GetBits(int bits)
//
// Desctiption: Extract number of bits from input stream.
// Returns: Right aligned bits. Last bit in Bit 0.
//
// bits - Number of bits to extract.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
int CBitStream::GetBits(int bits)
{
   int ReturnValue =0;

   while(bits >= 8)
   {
      ReturnValue <<= 8;
      if(m_BitIndex == 7)
      {
         ReturnValue |= (int)m_Data[m_ByteIndex] & 0x00ff;
         m_ByteIndex++;
      }
      else
      {
         // Get upper portion
         int Mask = 0x00fe << m_BitIndex;
         Mask = ~Mask;
         Mask &= 0x00ff;
         int tempValue = (int)m_Data[m_ByteIndex] & Mask;
         ReturnValue |= tempValue << (7 - m_BitIndex);
         m_ByteIndex++;
         // Get lower portion
         tempValue = (int)m_Data[m_ByteIndex] & 0x00ff;
         tempValue >>= (m_BitIndex + 1);
         ReturnValue |= tempValue;
      }
      bits -= 8;
   }

   if(bits >= (m_BitIndex + 1))
   {
      ReturnValue <<= m_BitIndex + 1;
      int Mask = 0x00fe << m_BitIndex;
      Mask = ~Mask;
      Mask &= 0x00ff;
      ReturnValue |= (int)m_Data[m_ByteIndex] & Mask;
      bits -= (m_BitIndex + 1);
      m_ByteIndex++;
      m_BitIndex = 7;
   }

   if(bits)
   {
      ReturnValue <<= bits;
      int Mask = 0x00fe << m_BitIndex;
      Mask = ~Mask;
      Mask &= 0x00ff;
      int tempValue = (int)m_Data[m_ByteIndex] & Mask;
      ReturnValue |= tempValue >> (m_BitIndex - bits + 1);
      m_BitIndex -= bits;
      if(m_BitIndex < 0)
         m_BitIndex = 7;
   }

   return ReturnValue;
}

/////////////////////////////////////////////////////////////////////////////
// RemainingBits()
//
// Desctiption: Remaining bits to exract.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
int CBitStream::RemainingBits()
{
   int Length = m_ByteIndex << 3;
   Length += 7 - m_BitIndex;
   Length = (m_SizeInBytes << 3) - Length;

   return Length;
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Desctiption: Reset input and output to initial state.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::Reset()
{
   SetInputBuffer(NULL, 0);
   ResetOutBuffers();
}

/////////////////////////////////////////////////////////////////////////////
// ResetOutBuffers()
//
// Desctiption: Reset output buffer to initial state.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::ResetOutBuffers()
{
   m_OutputBuffers.RemoveAll();
   m_OByteIndex = 0;
   m_OBitIndex = 0;
   m_ActiveOBuffer = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// Seek(int position)
//
// Desctiption: Reposition file pointer.
//
// position - offset from beginning of the active buffer.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CBitStream::Seek(int position)
{
	SetBufferIdx(position, 7);
}

/////////////////////////////////////////////////////////////////////////////
// SetInputBuffer(char *buffer, int bytesize)
//
// Desctiption: Set input buffer.
//
// buffer - Pointer to data in memory.
// bytesize - Size in bytes.
//
// TSTan   05 Nov 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::SetInputBuffer(char *buffer, int bytesize)
{
   m_Data = buffer;
   m_SizeInBytes = bytesize;
   m_ByteIndex = 0;
   m_BitIndex = 7;
}

/////////////////////////////////////////////////////////////////////////////
// SetBufferIdx(int byteidx, int bitidx)
//
// Desctiption: Set new index position.
//
// byteidx - index from beginning of current active buffer.
// bitidx - index from MSB.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CBitStream::SetBufferIdx(int byteidx, int bitidx)
{
   m_ByteIndex = byteidx;
   m_BitIndex = bitidx;
}

/////////////////////////////////////////////////////////////////////////////
// SkipBits(int bits)
//
// Desctiption: Get and discard bits.
//
// bits - number of bits to get and discard.
//
// TSTan   09 Sep 2000
/////////////////////////////////////////////////////////////////////////////
void CBitStream::SkipBits(int bits)
{
	int longwords = bits / 32;
	bits %= 32;

	while(longwords--)
		GetBits(32);

	GetBits(bits);
}

/////////////////////////////////////////////////////////////////////////////
// CFileBitStream
//
// Desctiption: Empty constructor
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
CFileBitStream::CFileBitStream(char *input, char *output)
{
	m_infile = NULL;
	m_outfile = NULL;

	m_inputbuffer = new char[BUFFER_LENGTH];

	if(input)
		SetInFile(input);
}

/////////////////////////////////////////////////////////////////////////////
// ~CFileBitStream
//
// Desctiption: Clean up.
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
CFileBitStream::~CFileBitStream()
{
   Reset();

	delete m_inputbuffer;
}

/////////////////////////////////////////////////////////////////////////////
// Reset()
//
// Desctiption: Clean up.
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
void CFileBitStream::Reset()
{
	CBitStream::Reset();
	if(m_infile)
		delete m_infile;
	m_infile = NULL;

	if(m_outfile)
		delete m_outfile;
	m_outfile = NULL;
}

/////////////////////////////////////////////////////////////////////////////
// SetInFile(char *filename)
//
// Desctiption: Open file for reading.
//
// filename - input filename.
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
void CFileBitStream::SetInFile(char *filename)
{
	if(m_infile)
		delete m_infile;

	m_infile = new ifstream(filename, ios::binary | ios::nocreate);
	m_infile->seekg(0, ios::end);
	m_infilesize = m_infile->tellg();
	m_totalsize = m_infilesize;
	m_infile->seekg(0, ios::beg);
}

/////////////////////////////////////////////////////////////////////////////
// GetBits(int bits)
//
// Desctiption: Get bits from file.
//
// bits - number of bits to read.
//
// TSTan   09 Sep 2002
/////////////////////////////////////////////////////////////////////////////
int CFileBitStream::GetBits(int bits)
{
	int data = 0;

   if(m_infile)
	{
		int remain = RemainingBits();

		// Get bits from current buffer pool
		if(remain < bits)
		{
			data = CBitStream::GetBits(remain);
			bits -= remain;
		}
		else
			return CBitStream::GetBits(bits);

		// Get next block if pool empty
		m_infile->read(m_inputbuffer, BUFFER_LENGTH);
		if(m_infilesize > BUFFER_LENGTH)
		{
			SetInputBuffer(m_inputbuffer, BUFFER_LENGTH);
			m_infilesize -= BUFFER_LENGTH;
		}
		else
		{
			SetInputBuffer(m_inputbuffer, int(m_infilesize));
			m_infilesize = 0;
		}

		if(RemainingBits() < bits)
		{
			remain = RemainingBits();
			data <<= remain;
			data |= CBitStream::GetBits(remain);
		}
		else
		{
			data <<= bits;
			data |= CBitStream::GetBits(bits);
		}
	}

	return data;
}

/////////////////////////////////////////////////////////////////////////////
// GetOffset()
//
// Desctiption: Return current file pointer offset relative to beginning.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
int CFileBitStream::GetOffset()
{
	int thisoffset;
	if(m_infilesize)
		thisoffset = m_totalsize - m_infilesize -	BUFFER_LENGTH + m_ByteIndex;
	else
		thisoffset = m_totalsize - (m_totalsize % BUFFER_LENGTH) + m_ByteIndex;

	return thisoffset;
}

/////////////////////////////////////////////////////////////////////////////
// RdBytesAvailable()
//
// Desctiption: Get total of remaining bytes in the file.
//
// TSTan   28 Sep 2002
/////////////////////////////////////////////////////////////////////////////
int CFileBitStream::RdBytesAvailable()
{
	return m_infilesize + (RemainingBits() + 7) / 8;
}

/////////////////////////////////////////////////////////////////////////////
// Seek(int position)
//
// Desctiption: Reposition file pointer.
//
// position - offset from beginning of the file.
//
// TSTan   02 Nov 2002
/////////////////////////////////////////////////////////////////////////////
void CFileBitStream::Seek(int position)
{
	if(m_infile)
	{
		int block_add = position - (position % BUFFER_LENGTH);
		m_infilesize = m_totalsize - block_add;

		m_infile->seekg(block_add, ios::beg);
		m_infile->clear();

		// Get new block
		m_infile->read(m_inputbuffer, BUFFER_LENGTH);
		if(m_infilesize > BUFFER_LENGTH)
		{
			SetInputBuffer(m_inputbuffer, BUFFER_LENGTH);
			m_infilesize -= BUFFER_LENGTH;
		}
		else
		{
			SetInputBuffer(m_inputbuffer, int(m_infilesize));
			m_infilesize = 0;
		}
		SetBufferIdx(position % BUFFER_LENGTH, 7);
	}
}
